/*
    TUIO Simulator - part of the reacTIVision project
    http://reactivision.sourceforge.net/

    Copyright (c) 2005-2008 Martin Kaltenbrunner <mkalten@iua.upf.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.awt.*;
import java.util.*;
import java.awt.geom.*;

public class Tangible {

	public int fiducial_id;
	public int session_id;
	public TangibleType type;
	
	public Shape geom;
	public Point2D center;
	public Color color;

	private boolean active;
	public float orientation = 0;

	private final float pi = (float)Math.PI;
	private final float halfPi = (float)(Math.PI/2);
	private final float doublePi = (float)(Math.PI*2);
	
	public float xspeed, yspeed, mspeed, maccel, rspeed, raccel;
	public Tangible previous = null;
	public Tangible next = null;

	private long lastTime;

	public Tangible(int s_id, int f_id, TangibleType type, boolean active, float xpos, float ypos, float angle) {
	
		this.session_id = s_id;
		this.fiducial_id = f_id;
		this.type = type;
		this.active = active;
		
		geom = type.getShape();
		center = type.center;
		color = type.color;

		float dx = xpos - (float)center.getX();
		float dy = ypos - (float)center.getY();
		translate(dx,dy);
	
		float dr =  angle;
		rotate(dr-orientation);
		
		lastTime = System.currentTimeMillis();

		xspeed = 0.0f;
		yspeed = 0.0f;
		mspeed = 0.0f;
		rspeed = 0.0f;
		maccel = 0.0f;
		raccel = 0.0f;
	}

	public final void translate(float dx, float dy) {

		AffineTransform  trans = AffineTransform.getTranslateInstance(dx,dy);
		geom = trans.createTransformedShape(geom);
		center = new Point2D.Float((float)(center.getX()+dx),(float)(center.getY()+dy));
		
		long currentTime = System.currentTimeMillis();
		long dt = currentTime - lastTime;
		float dist = (float)Math.sqrt(dx*dx+dy*dy);
		float new_speed  = dist/dt;
		this.xspeed = Math.abs(dx/dt);
		this.yspeed = Math.abs(dy/dt);
		this.maccel = (new_speed-mspeed)/dt;
		this.mspeed = new_speed;
		lastTime = currentTime;		
		
		if ((center.getX()<0) || (center.getY()<0)) deactivate();
	}

	public final void rotate(double theta) {

		orientation += theta;
		AffineTransform  trans = AffineTransform.getRotateInstance(theta,center.getX(),center.getY());
		geom = trans.createTransformedShape(geom);	

		long currentTime = System.currentTimeMillis();
		long dt = currentTime - lastTime;
		float new_speed  = (float)theta/dt;
		this.raccel = (new_speed-rspeed)/dt;
		this.rspeed = new_speed;
		lastTime = currentTime;		
	}
 
	public final Point getPosition() {
		return new Point((int)(center.getX()),(int)(center.getY()));
	}
	
	public Point getPointer() {

		int x = (int)(center.getX()-Math.cos(orientation-halfPi)*15);
		int y = (int)(center.getY()-Math.sin(orientation-halfPi)*15);
		return new Point(x,y);
	}
	
	public final float getAngle() {

		float angle =  (float)(orientation%doublePi);
		if (angle<0) angle += doublePi;
		return angle;
	}

	public final boolean containsPoint(int x, int y) {
		return geom.intersects(x,y,1,1);
	}

	public final void activate(int s_id) {

		active = true;
		session_id = s_id;
		
		lastTime = System.currentTimeMillis();
		xspeed = 0.0f;
		yspeed = 0.0f;
		mspeed = 0.0f;
		rspeed = 0.0f;
		maccel = 0.0f;
		raccel = 0.0f;
	}

	public final void deactivate() {

		active = false;
		
		lastTime = System.currentTimeMillis();
		xspeed = 0.0f;
		yspeed = 0.0f;
		mspeed = 0.0f;
		rspeed = 0.0f;
		maccel = 0.0f;
		raccel = 0.0f;		
	}
	
	public final boolean isActive() {

		return active;
	}
}
